# Symbolic math example equivalent to the spreadsheet provided for,
# Supporting material for Decision Modelling for Health Economic Evaluation 
# Chapter 2, http://www.gla.ac.uk/media/media_94488_en.xls
#
# Coding, Jacob Barhak Ph.D.
# http://sites.google.com/site/jacobbarhak/
# 19-Dec-2013
# 

###############################################################################
#                       Section A - Initialize 
# This section contains some definitions and required initializations
# No real computation will be found here

from __future__ import division
from sympy import symbols,Matrix, solve
from sympy.core.containers import Tuple


# Define some functions help substitute the results and display them. We will 
# need to repeat this several times - better have functions to reuse code.

def DisplayResults(ResultsList):
    "This function prints results to screen"
    print "Year   Population in State                  Life Years            Costs" 
    print "       A        B        C        D        No Disc  Disc      No Disc  Disc" 
    for (TimeEnum,TimeEntry) in enumerate(ResultsList):
        (PopulationInThisYearA, PopulationInThisYearB, PopulationInThisYearC, PopulationInThisYearD , LifeYears, LifeYearsDiscounted, CostsThisYear, CostsThisYearDiscounted, TotalCosts, TotalCostsDiscounted, TotalLifeYears, TotalLifeYearsDiscounted) = TimeEntry
        print "%3i, %0.5f, %0.5f, %0.5f, %0.5f,    %0.2f,   %0.2f,   %5.0f, %5.0f" % (TimeEnum+1, PopulationInThisYearA, PopulationInThisYearB, PopulationInThisYearC, PopulationInThisYearD, LifeYears, LifeYearsDiscounted, CostsThisYear, CostsThisYearDiscounted)
    print 
    print "                                     Sum:   %0.2f,   %0.2f,   %5.0f, %5.0f" % (TotalLifeYears, TotalLifeYearsDiscounted, TotalCosts, TotalCostsDiscounted)
    print 
    print "#"*75



def SubstituteResults(ResultsList,Substititions):
    "Substitutes all results lists with variables values"
    SubstitutedResults = ResultsList
    for (VariableToSubstitute, ValueToSubstitute) in Substititions: 
        SubstitutedResults = SubstitutedResults.subs(VariableToSubstitute, ValueToSubstitute)
    return SubstitutedResults


# Define transition probabilities From states A,B,C,D
tpA2B = symbols('tpA2B')
tpA2C = symbols('tpA2C')
tpA2D = symbols('tpA2D')
tpB2B = symbols('tpB2B')
tpB2C = symbols('tpB2C')
tpB2D = symbols('tpB2D')
tpC2C = symbols('tpC2C')
tpC2D = symbols('tpC2D')

# Define costs
dmca = symbols('dmca') # Direct medical costs associated with state A
dmcb = symbols('dmcb') # Direct medical costs associated with state B
dmcc = symbols('dmcc') # Direct medical costs associated with state C
ccca = symbols('ccca') # Community care costs associated with state A
cccb = symbols('cccb') # Community care costs associated with state B
cccc = symbols('cccc') # Community care costs associated with state C

# Drug Costs
cAZT = symbols('cAZT') # Zidovudine drug cost
cLam = symbols('cLam') # Lamivudine drug cost

# Other Parameters
RR = symbols('RR')  # Treatment effect (RR)
cDR =	symbols('cDR')   # Annual discount rate - costs (%)
oDR = symbols('oDR')   # Annual discount rate - benefits (%)

# Define diagonal of Markov Matrix to add up row values to 1
# We dont need these as inputs - these numbers can be deduced
tpA2A = 1 - tpA2B*RR - tpA2C*RR - tpA2D*RR
tpB2B = 1 - tpB2C*RR - tpB2D*RR
tpC2C = 1 - tpC2D*RR

MarkovMatrix = Matrix([ [tpA2A , tpA2B*RR, tpA2C*RR, tpA2D*RR],
                        [    0 ,    tpB2B, tpB2C*RR, tpB2D*RR],
                        [    0 ,        0,    tpC2C, tpC2D*RR],
                        [    0 ,        0,        0,        1] ])

InitialPop = Matrix([[1,0,0,0]])
AllResults = Tuple()
TotalCosts = 0
TotalCostsDiscounted = 0
TotalLifeYears = 0
TotalLifeYearsDiscounted = 0
PopulationInThisYear = InitialPop

###############################################################################
#                       Section B - Construct Expressions
# This section will construct the symbolic expressions. We still need to loop
# through the years to do this, just because of the discrete nature of the 
# problem at hand with all the constraints imposed. However, at the end of this
# loop we will have the expressions describing the problem rather than numeric
# results.

for YearValue in range(1,21):
    if YearValue <=2: 
        # only first two years have any risk reduction in any scenario
        ReplaceRR = RR
        GenericCostVector = Matrix([dmca+ccca+cAZT+cLam, dmcb+cccb+cAZT+cLam, dmcc+cccc+cAZT+cLam, 0])
    else:
        # after two years there is never any rik reduction
        ReplaceRR = 1
        GenericCostVector = Matrix([dmca+ccca+cAZT, dmcb+cccb+cAZT, dmcc+cccc+cAZT, 0])
    # calculate the population while substituting the RR 
    PopulationInThisYear = PopulationInThisYear * MarkovMatrix.subs(RR,ReplaceRR)
    [PopulationInThisYearA, PopulationInThisYearB, PopulationInThisYearC, PopulationInThisYearD] = PopulationInThisYear
    # The people in state D are considered for life years 
    LifeYears = 1 - (PopulationInThisYearD)
    # Apply discount
    LifeYearsDiscounted = LifeYears / ((1+oDR)**YearValue)
    # Calculate Costs
    CostsThisYear = PopulationInThisYear.dot(GenericCostVector)
    # Add the Discount
    CostsThisYearDiscounted = CostsThisYear / (1+cDR)**YearValue
    # Calculate totals up to this year
    TotalCosts = CostsThisYear + TotalCosts
    TotalCostsDiscounted = CostsThisYearDiscounted + TotalCostsDiscounted
    TotalLifeYears = LifeYears + TotalLifeYears
    TotalLifeYearsDiscounted = LifeYearsDiscounted + TotalLifeYearsDiscounted
    # Record all results in one vector with a list of elements each year
    AllResults = AllResults + Tuple((PopulationInThisYearA, PopulationInThisYearB, PopulationInThisYearC, PopulationInThisYearD , LifeYears, LifeYearsDiscounted, CostsThisYear, CostsThisYearDiscounted, TotalCosts, TotalCostsDiscounted, TotalLifeYears, TotalLifeYearsDiscounted))

# At this point all calculations have been made and all expressions for all 
# Variables for all years are defined as expressions.
###############################################################################
#                       Section C - Substitute and Display


# Define substitute results for the first scenario
Scenario1Inputs =    [    (tpA2B, (350 / (350 + 1384))),
                          (tpA2C, (116 / (116 + 1618))),
                          (tpA2D, (17 / (17 + 1717))),
                          (tpB2C, (512 / (512 + 746))),
                          (tpB2D, (15 / (15 + 1243))),
                          (tpC2D, (437 / (437 + 1312))),
                          (RR, 1.0), # No rist reduction first two years
                          (cDR, 0.06),
                          (oDR, 0),
                          (dmca, 1701),
                          (dmcb, 1774),
                          (dmcc, 6948),
                          (ccca, 1055),
                          (cccb, 1278),
                          (cccc, 2059),
                          (cAZT, 2278),
                          (cLam, 0)  # No extra costs first two years
                      ] 

   

# Define substitute results for the second scenario
Scenario2Inputs =    [    (tpA2B, (350 / (350 + 1384))),
                          (tpA2C, (116 / (116 + 1618))),
                          (tpA2D, (17 / (17 + 1717))),
                          (tpB2C, (512 / (512 + 746))),
                          (tpB2D, (15 / (15 + 1243))),
                          (tpC2D, (437 / (437 + 1312))),
                          (RR, 0.509),
                          (cDR, 0.06),
                          (oDR, 0),
                          (dmca, 1701),
                          (dmcb, 1774),
                          (dmcc, 6948),
                          (ccca, 1055),
                          (cccb, 1278),
                          (cccc, 2059),
                          (cAZT, 2278),
                          (cLam, 2086.5) 
                      ] 


# Now actually substitite the inputs in the expressions
ResultsScenario1 = SubstituteResults(AllResults,Scenario1Inputs)
print "#"*75
print
print  'Scenario A - MONOTHERAPY'
print
DisplayResults(ResultsScenario1)

ResultsScenario2 = SubstituteResults(AllResults,Scenario2Inputs)
print "#"*75
print
print  'Scenario B - COMBINATION THERAPY'
print
DisplayResults(ResultsScenario2)


###############################################################################
#                    Section D - Optimization Example
# This section will solve another problem that is not posed initially. The 
# problem is, what RR will cause undiscounted Life Years in Year 5 be 0.8?
# With other systems this will require writing a function dedicated to be 
# optimized. Using symbolic techniques it is quite easy to do using partial
# substitution.

OptScenarioInputs =   [   (tpA2B, (350 / (350 + 1384))),
                          (tpA2C, (116 / (116 + 1618))),
                          (tpA2D, (17 / (17 + 1717))),
                          (tpB2C, (512 / (512 + 746))),
                          (tpB2D, (15 / (15 + 1243))),
                          (tpC2D, (437 / (437 + 1312))),
                          # (RR, 1), Remove RR from substitutions is is unknown
                          (cDR, 0.06),
                          (oDR, 0),
                          (dmca, 1701),
                          (dmcb, 1774),
                          (dmcc, 6948),
                          (ccca, 1055),
                          (cccb, 1278),
                          (cccc, 2059),
                          (cAZT, 2278),
                          (cLam, 0)  # No extra costs first two years
                      ] 
# Now substitute all calculations                     
ExpresasionsScenario2 = SubstituteResults(AllResults,OptScenarioInputs)
# The expression obtained is a function of RR - Display it
(PopulationInThisYearA, PopulationInThisYearB, PopulationInThisYearC, PopulationInThisYearD , LifeYears, LifeYearsDiscounted, CostsThisYear, CostsThisYearDiscounted, TotalCosts, TotalCostsDiscounted, TotalLifeYears, TotalLifeYearsDiscounted) = ExpresasionsScenario2[4]
EquationToSolve = LifeYears - 0.8
print "#"*75
print  'Scenario C - Optimization problem not initially defined:'
print  'Problem: What RR will cause LY to be 0.8 in year 5 ? '
print
print 'Expression to solve is:'
print EquationToSolve
print 
# Simplify the expression
SimplifiedEquation = EquationToSolve.expand()
print 'This expression is simplified to:'
print  SimplifiedEquation
print 
Solution = solve(EquationToSolve)
print 'The solution for RR is:'
print Solution
print     
SelectedSolution = filter(lambda Entry:0<Entry<1, Solution)[0]
print 'Select the first valid RR solution between 0 an 1. It is the final answer:'
print SelectedSolution
print     
print 'Verify by substituting back into the expressions and showing results:'

ResultsSpecialScenario = SubstituteResults(ExpresasionsScenario2,[(RR, SelectedSolution)])
DisplayResults(ResultsSpecialScenario)

print     
print 'You should be able to see LY = 0.8 in year 5 !!!'
